// ====================================================
//
// Copyright (c) 2001 Sean Wilson. All Rights Reserved.
//
// ====================================================

import java.awt.image.*;
import java.awt.*;
import java.applet.*;

public class StereogramApplet extends Applet implements Runnable
{
	/** Thread used for generating stereogram. */
	private Thread updateThread;
	/** Finished stereogram image. */
	private Image stereogramImage;
	/** Applet status message. */
	private Label messageLabel;

	/** Default minimum pixel pair seperation in stereogram. */
	private final int DEFAULT_MIN_SEPERATION = 100;
	/** Default maximum pixel pair seperation difference from minimum. */
	private final int DEFAULT_SEPERATION_RANGE = 20;
	
	/** Generates a stereogram image. 
	 *  @param  minSeperation   minimum pixel pair seperation.
	 *  @param  seperationRange maximum pixel pair seperation difference from minimum.
	 *  @param  depthMap        depth-map image to use, where white is closest depth and black is furthest away.
	 *  @param  pattern         image to be used as tiled pattern. If this is <code>null</code>, random colors are used.
	 *  @return                 stereogram image same size as provided depth-map.
	 */
	public Image generateStereogram(Dimension size, int minSeperation, int seperationRange, Image depthMap, Image pattern)
	{
		// Store depth-map width and height
		int depthMapWidth  = depthMap.getWidth(null);
		int depthMapHeight = depthMap.getHeight(null);
		
		// Store pattern width and height if there is one
		int patternWidth  = (pattern == null ? 0 : pattern.getWidth (null));
		int patternHeight = (pattern == null ? 0 : pattern.getHeight(null));

		// Arrays to hold pixels for stereogram, depth-map and pattern images
		int[] patternPixel    = (pattern == null ? null : new int[patternWidth * patternHeight]);
		int[] stereogramPixel = new int[depthMapWidth  * depthMapHeight];
		int[] depthPixel      = new int[depthMapWidth  * depthMapHeight];

		// Pixel grabber for pattern and depth-map images
		PixelGrabber patternPixelGrabber = (pattern == null ? null : new PixelGrabber(pattern,  0, 0, patternWidth, patternHeight, patternPixel, 0, patternWidth));
		PixelGrabber depthPixelGrabber   = new PixelGrabber(depthMap, 0, 0, depthMapWidth,  depthMapHeight,  depthPixel,   0, depthMapWidth);

		try
		{
			// If there is a pattern image
			if (pattern != null)
			{
				// Get pixels from pattern image
				patternPixelGrabber.grabPixels();
			}
			
			// Get pixels from depth-map image
			depthPixelGrabber.grabPixels();
		}
		catch (InterruptedException e)
		{
			throw new RuntimeException("interrupted waiting to grab pixels");
		}

		// Maximum color value depth-map pixels can have
		final int MAX_COLOR_VALUE = 255 * 3;
		// Current pixel in depth-map and stereogram pixel arrays
		int pixelOffset = 0;

		// For each row
		for (int y = 0; y < depthMapHeight; ++y)
		{
			// For each column
			for (int x = 0; x < depthMapWidth; ++x)
			{
				// Get depth-map pixel color
				Color color = new Color(depthPixel[pixelOffset]);
				// Get pixel color value
				int colorValue = color.getRed() + color.getGreen() + color.getBlue();
				// Convert pixel color value to depth value bewteen 0.0 (furthest depth) and 1.0 (closest depth)
				float depthPercent = 1 - (float)colorValue / MAX_COLOR_VALUE;

				// Calculate distance other pixel of this matching pair is from this pixel, whose color value should be made the same
				int seperation = (int)(minSeperation + depthPercent * seperationRange);

				// If matching pixel is inside stereogram image
				if (seperation >= 0 && x - seperation >= 0)
				{
					// Set stereogram pixel to same color as matching pixel
					stereogramPixel[pixelOffset] = stereogramPixel[pixelOffset - seperation];
				}
				else
				{
					// If there is a pattern image
					if (pattern == null)
					{
						// Set stereogram pixel to random color
						stereogramPixel[pixelOffset] = new Color((float)Math.random(), (float)Math.random(), (float)Math.random()).getRGB();
					}
					else
					{
						// Set stereogram pixel to same color as if pattern were tiled here
						stereogramPixel[pixelOffset] = patternPixel[(x % patternWidth) + (y % patternHeight) * patternWidth];
					}
				}

				// Move to next pixel
				++pixelOffset;
			}
		}

		// Return stereogram image
		return createImage(new MemoryImageSource(depthMapWidth, depthMapHeight, stereogramPixel, 0, depthMapWidth));
	}

	/** Creates a resized image.
	 *  @param image         image to resize from.
	 *  @param newDimensions new dimensions resized image will have.
	 *  @return              resized image.
	 */
	public Image resizeImage(Image image, Dimension newDimensions)
	{
		// Create image with new size
		Image resizedImage = createImage(newDimensions.width, newDimensions.height);
		// Scale original image to get resized image
		resizedImage.getGraphics().drawImage(image, 0, 0, newDimensions.width, newDimensions.height, null);
		// Return resized image
		return resizedImage;
	}

	public void init()
	{
		// Use border layout
		setLayout(new BorderLayout());
	}

	public void start()
	{
		// Create update thread
		updateThread = new Thread(this);
		// Start update thread
		updateThread.start();
	}

	public void stop()
	{
		// Stop update thread
		updateThread = null;
	}

	public void destroy()
	{
		// Stop update thread
		updateThread = null;
	}

	public void showMessage(String message)
	{
		// If message label already added
		if (messageLabel != null)
		{
			// Remove old message label
			remove(messageLabel);
		}

		// Store new message label
		messageLabel = new Label(message, Label.CENTER);
		// Add label to center of applet
		add(messageLabel, BorderLayout.CENTER);
		// Show message
		repaint();
	}

	public void run()
	{
		// If stereogram image already generated
		if (stereogramImage != null)
		{
			// Stop update thread
			updateThread = null;
			return;
		}

		// Media tracker to wait for images loading
		MediaTracker mediaTracker = new MediaTracker(this);
		
		// Get minimum seperation parameter string
		String minSeperationString = getParameter("MinSeperation");
		// Use default minimum seperation if parameter was not there
		int minSeperation = (minSeperationString == null ? DEFAULT_MIN_SEPERATION : Integer.parseInt(minSeperationString));

		// Get seperation range parameter string
		String seperationRangeString = getParameter("SeperationRange");
		// Use default seperation range if parameter was not there
		int seperationRange = (seperationRangeString == null ? DEFAULT_SEPERATION_RANGE : Integer.parseInt(seperationRangeString));

		// Get depth-map image parameter string
		String depthImageName = getParameter("DepthImage");
		// If parameter was missing
		if (depthImageName == null)
		{
			throw new RuntimeException("\"DepthImage\" applet parameter is missing.");
		}
		
		// ID for depth-map image
		final int DEPTH_IMAGE_ID = 0;	
		// Get depth-map image
		Image depthMap = getImage(getDocumentBase(), depthImageName);
		// Add image to media tracker
		mediaTracker.addImage(depthMap, DEPTH_IMAGE_ID);

		// ID for pattern image
		final int PATTERN_IMAGE_ID = 1;
		// Pattern image
		Image pattern = null;
		// Get pattern image parameter string
		String patternImageName = getParameter("PatternImage");

		// If parameter was there
		if (patternImageName != null)
		{
			// Get pattern image
			pattern = getImage(getDocumentBase(), patternImageName);
			// Add image to media tracker
			mediaTracker.addImage(pattern, PATTERN_IMAGE_ID);
		}

		try
		{
			// If pattern parameter was supplied
			if (pattern != null)
			{
				// Wait for pattern image to load
				showMessage("Loading pattern image...");	
				mediaTracker.waitForID(PATTERN_IMAGE_ID);

				// If error loading pattern image
				if (mediaTracker.isErrorID(PATTERN_IMAGE_ID))
				{
					throw new RuntimeException("\"PatternImage\" applet parameter points to image that cannot be loaded.");
				}
			}

			// Wait for depth-map image to load
			showMessage("Loading depth-map image...");
			mediaTracker.waitForID(DEPTH_IMAGE_ID);

			// If error loading depth-map image
			if (mediaTracker.isErrorID(DEPTH_IMAGE_ID))
			{
				throw new RuntimeException("\"DepthImage\" applet parameter points to image that cannot be loaded.");
			}
		}
		catch (InterruptedException e)
		{
		}

		// Resize depth-map to fill applet window
		depthMap = resizeImage(depthMap, getSize());

		// Get stereogram image
		showMessage("Generating stereogram...");
		stereogramImage = generateStereogram(getSize(), minSeperation, seperationRange, depthMap, pattern);

		// Remove message label
		remove(messageLabel);
		// Show stereogram image
		repaint();

		// Stop update thread
		updateThread = null;
	}

	public void update(Graphics g)
	{
		// Paint components
		paintAll(g);
		// Paint stereogram
		paint(g);
	}

	public void paint(Graphics g)
	{
		// If stereogram image has been created
		if (stereogramImage != null)
		{
			// Draw stereogram image
			g.drawImage(stereogramImage, 0, 0, null);
		}
	}

	public String getAppletInfo()
	{
		return "Stereogram generator - Copyright (c) 2001 Sean Wilson. All Rights Reserved.";
	}

	public String[][] getParameterInfo()
	{
		String[][] info = { {"DepthImage",      "image filename",        "Image to be used as depth map where white is closest depth and black is furthest away."                                           },
		                    {"PatternImage",    "image filename",        "Image to be used as tiled pattern. If unprovided, random colors will be used instead."                                            },
		                    {"MinSeperation",   "positive whole number", "Minimum pixel pair seperation (closest depth). If unprovided, " + DEFAULT_MIN_SEPERATION + " is used."                            },
		                    {"SeperationRange", "positive whole number", "Maximum pixel pair seperation (furthest depth) difference from minimum. If unprovided, " + DEFAULT_SEPERATION_RANGE + " is used." },
		                  };
		
		return info;
	}
}